#!/usr/bin/env python

# Licensed under GPL, written by John Stephenson

# Extensive thanks to Giancarlo Fringuello (author of Kopete plasmoid) 
# and to meek (author of Pidgin screenlet) Thanks also to Chris Russell for
# bringing back the transparency

import dbus
import gobject
import os
import re
from dbus.mainloop.glib import DBusGMainLoop

from PyQt4.QtCore import *
from PyQt4.QtGui import *
from PyKDE4.plasma import Plasma
from PyKDE4 import plasmascript
from PyKDE4.kdeui import *
from PyKDE4.kdecore import i18n

class Buddy:
	def __init__ (self, id, alias, name=None, status=None, event_status=None,
			status_text=None, protocol=None, iconQImage=None, group=None):
		self.id = id
		self.alias = alias
		self.name = name
		self.group = group
		self.status = status
		self.event_status = event_status
		self.status_text = status_text
		self.protocol = protocol
		self.iconQImage = iconQImage


class Group:
	def __init__ (self, id, name, buddies=None, online_buddies=None):
		self.id = id
		self.name = name
		if buddies: self.buddies = buddies
		else: self.buddies = {}
		if online_buddies: self.online_buddies = online_buddies
		else: self.online_buddies = {}

from pidginitem import PidginItem

import sys
from stat import *
import os 

class PidginPlasmoid(plasmascript.Applet):
	pidgin = None
	show_offline = False
	show_groups = True
	group_by_groups = True
	group_by_account = False
	group_by_online_offline = True
	notify_time = 10000 # 10 seconds
	purple_path = "/im/pidgin/purple/PurpleObject"
	purple_service = "im.pidgin.purple.PurpleService"
	purple_interface = "im.pidgin.purple.PurpleInterface"
	purple_xmlfile = os.environ["HOME"]+"/.purple/blist.xml"
	
	displayfilter = ""

	# Buddies indexed by BuddyId
	buddies = {}
	# Groups indexed by GroupId
	groups = {}

	signonoff_timeouts = {}
	recentSignedOnBuddies = {}
	recentSignedOffBuddies = {}
	collapsedGroups = {}
	refresh_timeout = None

	def getBuddyGroupNameAndId(self,BuddyId):
		GroupName = None
		GroupId = None
		isOnline = self.pidgin.PurpleBuddyIsOnline(BuddyId)
		if self.show_groups and self.group_by_groups:
			GroupId = self.pidgin.PurpleBuddyGetGroup(BuddyId)
			GroupName = self.pidgin.PurpleGroupGetName(GroupId)
		elif self.show_groups and self.group_by_account:
			GroupId = self.pidgin.PurpleBuddyGetAccount(BuddyId)
			actAlias = self.pidgin.PurpleAccountGetAlias(GroupId)
			GroupName = self.pidgin.PurpleAccountGetProtocolName(GroupId)
			if actAlias:
				GroupName = actAlias+" ("+GroupName+")"
		elif self.show_groups and self.group_by_online_offline:
			if isOnline:
				GroupId = -1
				GroupName = "Online"
			else:
				GroupId = -2
				GroupName = "Offline"
		else:
			GroupId = -3
			GroupName = "Contacts"
			
		return (GroupName, GroupId)
	
	def getBuddyStatus(self,buddyId):
		presence = self.pidgin.PurpleBuddyGetPresence(buddyId)
		isOnline = self.pidgin.PurplePresenceIsOnline(presence)
		isAvail = self.pidgin.PurplePresenceIsAvailable(presence)
		isIdle = self.pidgin.PurplePresenceIsIdle(presence)
		status = self.pidgin.PurplePresenceGetActiveStatus(presence)
		status_text =  self.pidgin.PurpleStatusGetName(status)
		custom_status_text =  self.pidgin.PurpleStatusGetAttrString(status, "message")
	
		if not isOnline: status = "offline"
		elif isIdle: status = "idle"
		elif status_text=="Do Not Disturb": status = "busy"
		elif status_text=="Busy": status = "busy"
		elif not isAvail: status = "away"
		elif isOnline and isAvail: status = "online"
		if isIdle and (status_text != "Idle"): status_text += "/Idle"
		if status_text=="Available": status_text = ""
	
		if custom_status_text:
			status_text = re.sub(r'<[^>]*?>', '', custom_status_text)
		return (status, status_text)
	
	def setBuddyIconImage(self,buddy, size):
		buddyId = buddy.id
		iconId = self.pidgin.PurpleBuddyGetIcon(buddyId)
		if iconId:
			iconPath = self.pidgin.PurpleBuddyIconGetFullPath(iconId)
			if iconPath and os.path.exists(iconPath) and not iconPath.endswith('.icon'): 
				buddy.iconQImage = iconPath
	
	def addBuddyAndGroup(self,BuddyId, GroupId, GroupName):
	 	
		alias = self.pidgin.PurpleBuddyGetAlias(BuddyId)
		group_obj = None
		if GroupId in self.groups:
			group_obj = self.groups[GroupId]
		else:
			group_obj = Group(GroupId, GroupName)
	
		buddy_obj = Buddy(BuddyId, alias)
		self.buddies[BuddyId] = buddy_obj
		self.groups[GroupId] = group_obj
	
		buddy_obj.name = self.pidgin.PurpleBuddyGetName(BuddyId)
		buddy_obj.group = group_obj
		(buddy_obj.status, buddy_obj.status_text) = self.getBuddyStatus(BuddyId)
		buddy_obj.protocol = self.pidgin.PurpleAccountGetProtocolName(
			self.pidgin.PurpleBuddyGetAccount(BuddyId))
		self.setBuddyIconImage(buddy_obj, 32)
		group_obj.buddies[BuddyId] = buddy_obj
		if buddy_obj.status != "offline":
			group_obj.online_buddies[BuddyId] = buddy_obj
	
	def addBuddy(self,BuddyId):
		(GroupName, GroupId) = self.getBuddyGroupNameAndId(BuddyId)
		if GroupName and GroupId:
			self.addBuddyAndGroup(BuddyId, GroupId, GroupName)
			return True
	 	return False

	def removeBuddy(self, buddyId):
		if buddyId in self.buddies:
			group = self.buddies[buddyId].group
			del group.buddies[buddyId]
			del self.buddies[buddyId]
			if len(group.buddies) == 0:
				del self.groups[group.id]
			return True
	 	return False
		
	def getGroupsAndBuddies(self):
		if self.pidgin:
			print "Refresh groups and buddies (offline: ",self.show_offline,")"
			text = ""
			self.buddies.clear()
			self.groups.clear()
			# traverse through the buddy list
			node = self.pidgin.PurpleBlistGetRoot()
			while node != 0:
				if self.pidgin.PurpleBlistNodeIsBuddy(node):
					BuddyId = node
					actId = self.pidgin.PurpleBuddyGetAccount(BuddyId)
					if self.pidgin.PurpleAccountIsConnected(actId):
						isOnline = self.pidgin.PurpleBuddyIsOnline(BuddyId)
						if isOnline or self.show_offline:
							self.addBuddy(BuddyId)
				node = self.pidgin.PurpleBlistNodeNext(node, 0)

	def removeRecentSignOnState(self, buddy):
		print "Removing recentSignOn State for ",buddy
		del self.recentSignedOnBuddies[buddy]
		if buddy in self.signonoff_timeouts:
			gobject.source_remove(self.signonoff_timeouts[buddy])
			del self.signonoff_timeouts[buddy]
		self.drawContacts()

	def removeRecentSignOffState(self, buddy):
		print "Removing recentSignOff State for ",buddy
		del self.recentSignedOffBuddies[buddy]
		if buddy in self.signonoff_timeouts:
			gobject.source_remove(self.signonoff_timeouts[buddy])
			del self.signonoff_timeouts[buddy]
		if self.removeBuddy(buddy):
			self.drawContacts()
		
	# signal binding to Pidgin	
	def ReceivedImMsg(self,account, sender, message, conversation, flags):
		print "ReceivedImMsg: ",sender, "said:", message
		self.getGroupsAndBuddies()
		self.drawContacts()
	
	def SignedOn(self,connection):
		print "SignedOn: ",connection
		self.getGroupsAndBuddies()
		self.drawContacts()
	
	def SignedOff(self,connection):
		print "SignedOff: ",connection
		self.getGroupsAndBuddies()
		self.drawContacts()
	
	def BuddySignedOn(self,buddy):
		print "BuddySignedOn: ",buddy
		if buddy in self.recentSignedOffBuddies:
			del self.recentSignedOffBuddies[buddy]
		if buddy in self.signonoff_timeouts:
			gobject.source_remove(self.signonoff_timeouts[buddy])
			del self.signonoff_timeouts[buddy]
		self.recentSignedOnBuddies[buddy] = 1
		self.signonoff_timeouts[buddy] = gobject.timeout_add(
			self.notify_time,
			self.removeRecentSignOnState, buddy)
		if self.addBuddy(buddy):
			self.drawContacts()
	
	def BuddySignedOff(self,buddy):
		print "BuddySignedOff: ",buddy
		(self.buddies[buddy].status, self.buddies[buddy].status_text) = self.getBuddyStatus(buddy)
		if buddy in self.recentSignedOnBuddies:
			del self.recentSignedOnBuddies[buddy]
		if buddy in self.signonoff_timeouts:
			gobject.source_remove(self.signonoff_timeouts[buddy])
			del self.signonoff_timeouts[buddy]
		self.recentSignedOffBuddies[buddy] = 1
		self.signonoff_timeouts[buddy] = gobject.timeout_add(
				self.notify_time,
				self.removeRecentSignOffState, buddy)
		self.drawContacts()
	
	def BuddyStatusChanged(self,buddy,old_status,status):
		print "BuddyStatusChanged: ",buddy," old status: ",old_status," new status: ",status
		(self.buddies[buddy].status, self.buddies[buddy].status_text) = self.getBuddyStatus(buddy)
		#self.getGroupsAndBuddies()
		self.drawContacts()
	
	def BuddyIconChanged(self,buddy):
		print "BuddyIconChanged: ",buddy
		self.getGroupsAndBuddies()
		self.drawContacts()
	
	#def BuddyTyping(self,account,name):
		#print "BuddyTyping: ",account, name
		#self.getGroupsAndBuddies()
		#self.drawContacts()

	def __init__(self,parent,args=None):
		plasmascript.Applet.__init__(self,parent)

	def bindSignals(self):
		if self.pidgin:
			print 'Setting up signals'
			self.bus.add_signal_receiver(self.ReceivedImMsg,dbus_interface="im.pidgin.purple.PurpleInterface",
				signal_name="ReceivedImMsg")
			self.bus.add_signal_receiver(self.SignedOn,dbus_interface="im.pidgin.purple.PurpleInterface",
				signal_name="SignedOn")
			self.bus.add_signal_receiver(self.SignedOff,dbus_interface="im.pidgin.purple.PurpleInterface",
				signal_name="SignedOff")
			self.bus.add_signal_receiver(self.BuddySignedOn,dbus_interface="im.pidgin.purple.PurpleInterface",
				signal_name="BuddySignedOn")
			self.bus.add_signal_receiver(self.BuddySignedOff,dbus_interface="im.pidgin.purple.PurpleInterface",
				signal_name="BuddySignedOff")
			self.bus.add_signal_receiver(self.BuddyStatusChanged,dbus_interface="im.pidgin.purple.PurpleInterface",
				signal_name="BuddyStatusChanged")
			self.bus.add_signal_receiver(self.BuddyIconChanged,dbus_interface="im.pidgin.purple.PurpleInterface",
				signal_name="BuddyIconChanged")
			#self.bus.add_signal_receiver(self.BuddyTyping,dbus_interface="im.pidgin.purple.PurpleInterface",
			#	signal_name="BuddyTyping")
		else:
			print "error: bind signals!"

	def onDBusServiceChange(self, service, arg2, arg3):
		if service == self.purple_service:
			if arg2 == "":
				print "Pidgin Started! :)"
				self.pidgin = self.getPidginDbus()
				self.drawContacts()
			if arg3 == "":
				print "Pidgin closed... :("
				self.recentSignedOnBuddies.clear()
				self.recentSignedOffBuddies.clear()
				self.pidgin = None
				self.drawContacts()
	
	def getPidginDbus(self):
		try:
			obj = self.bus.get_object(
			    self.purple_service, self.purple_path)
			self.pidgin = dbus.Interface(obj, self.purple_interface)
		except:
			print 'The Pidgin Session is not present'
			self.pidgin = None

		if self.pidgin:
			self.getGroupsAndBuddies()
			self.bindSignals()
			# Update after some timeout or it doesn't work properly
			print "Found Pidgin.."

		return self.pidgin
 

	def __init__(self,parent):
		plasmascript.Applet.__init__(self, parent)
		
	def init(self):
		self.setHasConfigurationInterface(True)
		self.resize(300, 900)
		self.setAspectRatioMode(Plasma.IgnoreAspectRatio)
		dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

		self.bus = dbus.SessionBus()
		if self.bus:
			self.bus.add_signal_receiver(self.onDBusServiceChange,
				dbus_interface="org.freedesktop.DBus",signal_name="NameOwnerChanged")
			self.pidgin = self.getPidginDbus()
		else:
			print 'dbus is not present!'
		
		self.theme = Plasma.Svg(self)
		self.theme.setImagePath("widgets/background")
		self.setBackgroundHints(Plasma.Applet.DefaultBackground)
		
		self.contactsList = QTreeWidget()
		# should set background color of self.contactsList - but how?
		self.contactsList.setColumnCount(2)
		self.contactsList.header().setResizeMode(1, QHeaderView.Stretch)
		self.contactsList.header().setResizeMode(0, QHeaderView.ResizeToContents)
		
		p = QPalette();
		p.setColor(QPalette.Base,QColor(255,255,255,128));
		p.setColor(QPalette.Text,QColor(0,0,0,255));
		self.contactsList.setPalette(p);
		
		self.contactsList.header().hide()
		self.contactsList.setIconSize(QSize(32, 32))
		self.contactsList.setRootIsDecorated(False)
		
		self.contactsList.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
		self.contactsList.setVerticalScrollMode(QAbstractItemView.ScrollPerItem)
		
		self.mainLayout = QGridLayout()
		self.mainLayout.setMargin(0)
		self.contactsList.itemDoubleClicked.connect(self.clicked)

		self.searchBox = KLineEdit()
		self.searchBox.setClearButtonShown(True)
		self.searchBox.setClickMessage("Search all contacts")
		self.searchBox.textChanged.connect(self.filterchanged)

		p = QPalette();
		p.setColor(QPalette.Base,QColor("transparent"));
		p.setColor(QPalette.Text,QColor(0,0,0,128));
		self.searchBox.setPalette(p);
		
		self.mainLayout.addWidget(self.searchBox)
		self.mainLayout.addWidget(self.contactsList)
	
		groupBox = Plasma.GroupBox()
		groupBox.nativeWidget().setFlat(True)
		groupBox.nativeWidget().setLayout(self.mainLayout)
		
		self.layout = QGraphicsLinearLayout(self.applet)
		self.layout.setOrientation(Qt.Vertical)
		self.layout.addItem(groupBox)
		
		self.drawContacts()

	def delayed(self, function, start_flag=True):
		if start_flag and not self.refresh_timeout:
			print "First up! Will draw in 1.5s"
			self.refresh_timeout = gobject.timeout_add(1500, 
					self.delayed, function, False)
		elif start_flag and self.refresh_timeout:
			print "Ignoring draw request, already made a request"
			None
		elif not start_flag:
			function()
			self.refresh_timeout = None	

	def drawContacts(self):
		self.delayed(self.drawContactsNow)
		
	def drawContactsNow(self):
		# this might be better to have it to queue for 250ms then run the 
		# real function to avoid excessive calls that come from DBus being
		# flooded with Pidgin events
		items = []
		if self.pidgin:
			print 'Drawing contacts'
			for GroupId in self.groups:
				if self.show_groups:
					items.append(
						PidginItem(
							self,
							GroupId,
							0,
							self.groups[GroupId].name,
							len(self.groups[GroupId].buddies)
						)
					)
				if GroupId in self.collapsedGroups:
					print "not showing ",self.groups[GroupId].name
				else:
					for BuddyId in self.groups[GroupId].buddies:
						# search alias & name for display filter
						searchtext = self.buddies[BuddyId].alias.lower() + " " + self.buddies[BuddyId].name.lower()
						if searchtext.find(str(self.displayfilter).lower()) != -1:
							# determine if they're a recently signed in/off contact
							if BuddyId in self.recentSignedOffBuddies: recent = -1
							elif BuddyId in self.recentSignedOnBuddies:	recent = 1
							else:	recent = 0
							items.append(
								PidginItem(
									self,
									GroupId,
									BuddyId,
									self.buddies[BuddyId].alias,
									self.buddies[BuddyId].status_text,
									self.buddies[BuddyId].name,
									self.buddies[BuddyId].status,
									self.buddies[BuddyId].iconQImage,
									recent
								)
							)
		else:
			items.append(
				PidginItem(
					self,
					0,
					1,
					"Pidgin not running",
					"Double-click to start",
					"",
					0,
					'/usr/share/icons/hicolor/48x48/apps/pidgin.png',
					-2)
			)
			print "Can't draw contacts - pidgin not running"

		self.contactsList.clear()
		self.contactsList.insertTopLevelItems(0, items)
	
	def filterchanged(self):
		self.displayfilter = self.searchBox.text()
		# if we're clearing the filter, go back to just online buddies
		if self.displayfilter == "":
			self.show_offline = False
			self.getGroupsAndBuddies()
		# if we're searching and we haven't shown offline buddies, show them
		elif self.show_offline == False:
			self.show_offline = True
			self.getGroupsAndBuddies()
		self.drawContacts()
	
	def clicked(self, item, row):
		if item.group == 0:
			# this doesn't seem to work:
			item.name = "Starting pidgin..."
			item.alias = "Please standby"
			print "launching pidgin"
			os.system("pidgin")
		elif item.buddy == 0:
			print "collapse/expand group: "+item.alias
			if item.group in self.collapsedGroups:
				del self.collapsedGroups[item.group]
			else:
				self.collapsedGroups[item.group] = 1
			self.drawContacts()
		else:
			print "start conversation with: "+item.alias
			self.pidgin.PurpleConversationPresent(self.pidgin.PurpleConversationNew(1, 
				self.pidgin.PurpleBuddyGetAccount(int(item.buddy)), 
				self.pidgin.PurpleBuddyGetName(int(item.buddy))))
			self.contactsList.setItemSelected(item,False)

def CreateApplet(parent):
	return PidginPlasmoid(parent) 

